package com.jeffknecht.dynadao.impl;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sql.DataSource;

import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.dao.OptimisticLockingFailureException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.BeanPropertySqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.api.DynaDao;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;
import com.jeffknecht.dynadao.api.Generator;


/**
 * Dynamic data access object.  Provides the ability to 
 * query and updated database tables based on annotations in the
 * DTO class definition.
 * 
 * Note that the intention here is not to replace JPA or Hibernate.
 * These frameworks are very good at what they do, but introduce some
 * interesting complexities with respect to session management.  Rather,
 * the intention of this library is to simplify the 80% usage scenario
 * of simple CRUD operations against a single table.
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoImpl implements DynaDao {
	protected static final HashMap<Class<? extends Generator>, Generator> generators = new HashMap<Class<? extends Generator>, Generator>();
	protected NamedParameterJdbcTemplate namedParameterjdbcTemplate;
	protected JdbcTemplate jdbcTemplate;
	
	public DynaDaoImpl(DataSource ds) {
		this.jdbcTemplate = new JdbcTemplate(ds);
		this.namedParameterjdbcTemplate = new NamedParameterJdbcTemplate(ds);
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#findAll(java.lang.Class)
	 */
	public <T> List<T> findAll(Class<T> clazz) {
		String sql = baseSelectSQL(clazz);
		return jdbcTemplate.query(sql, new DynaRowMapper<T>(clazz), (Object[]) null);
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.HashMap)
	 */
	public <T> List<T> find(Class<T> clazz, final String where, HashMap<String, ?> params) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Pattern p = Pattern.compile("\\{[^\\{\\}]*\\}");	// match any text surrounded by {curly braces}
															// except ignore curly braces inside of curly braces
		
		String targetWhere = where.trim();
		Matcher matcher = p.matcher(targetWhere);
		while (matcher.find()) {
			String toReplace = matcher.group();
			String fieldName = toReplace.substring(1, toReplace.length() - 1);
			targetWhere = targetWhere.replace(toReplace, resolveColumnName(dtoHelper, fieldName));
			matcher = p.matcher(targetWhere);
		}
		
		sql.append(" ");
		if (!targetWhere.startsWith("where ")) {
			sql.append("where ");
		}
		sql.append(targetWhere);
		return namedParameterjdbcTemplate.query(sql.toString(), params, new DynaRowMapper<T>(clazz));
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.HashMap)
	 */
	public <T> T findOne(Class<T> clazz, final String where, HashMap<String, ?> params) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Pattern p = Pattern.compile("\\{[^\\{\\}]*\\}");	// match any text surrounded by {curly braces}
															// except ignore curly braces inside of curly braces
		
		String targetWhere = where.trim();
		Matcher matcher = p.matcher(targetWhere);
		while (matcher.find()) {
			String toReplace = matcher.group();
			String fieldName = toReplace.substring(1, toReplace.length() - 1);
			targetWhere = targetWhere.replace(toReplace, resolveColumnName(dtoHelper, fieldName));
			matcher = p.matcher(targetWhere);
		}
		
		sql.append(" ");
		if (!targetWhere.startsWith("where ")) {
			sql.append("where ");
		}
		sql.append(targetWhere);
		try {
			return namedParameterjdbcTemplate.queryForObject(sql.toString(), params, new DynaRowMapper<T>(clazz));
		} catch (EmptyResultDataAccessException ignored) {
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#insert(java.lang.Object)
	 */
	public int insert(Object object) {
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		String sql = insertSQL(object.getClass());

		//generate the column values
		for (String fieldName : dtoHelper.fieldNames()) {
			Field field = dtoHelper.field(fieldName);
			Column column = field.getAnnotation(Column.class);
			Class<? extends Generator> generatorClass = column.generator();
			if (column.generateOnInsert() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
				try {
					Generator generator = generators.get(generatorClass);
					if (generator == null) {
						generator = generatorClass.newInstance();
						generators.put(generatorClass, generator);
					}
					boolean accessible = field.isAccessible();
					field.setAccessible(true);
					generator.generateValue(field, object, Generator.Operation.INSERT);
					field.setAccessible(accessible);
				} catch (InstantiationException e) {
					throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
				} catch (IllegalAccessException e) {
					throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
				}
			}
		}
		
		if (dtoHelper.autoIncrementFieldName() == null) {
			return namedParameterjdbcTemplate.update(sql, new BeanPropertySqlParameterSource(object));
		} else {
			String fieldName = dtoHelper.autoIncrementFieldName();
			GeneratedKeyHolder keyHolder = new GeneratedKeyHolder();
			int returnValue = namedParameterjdbcTemplate.update(sql, new BeanPropertySqlParameterSource(object), keyHolder);
			Number key = keyHolder.getKey();
			Field field = dtoHelper.field(fieldName);
			boolean accessible = field.isAccessible();
			field.setAccessible(true);
			try {
				field.setInt(object, key.intValue());
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access on field " + field, e);
			}
			field.setAccessible(accessible);
			return returnValue;
		}
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#update(java.lang.Object)
	 */
	public int update(Object object) {
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
		Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
		if (primaryKeyFieldNames.size() == 0) {
			throw new DynamicDaoDataAccessException("Unable to uniquely identify object " + object + ", because it has no primary key fields");
		}
		StringBuilder sqlBuilder = new StringBuilder(baseUpdateSQL(object.getClass()));
		sqlBuilder.append(" where ");
		int whereCounter = 0;
		for(String fieldName : primaryKeyFieldNames) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName);
			whereCounter++;
		}
		for(String fieldName : dtoHelper.optimisticLockingFieldNames()) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName + "_optlock");
			Object fieldValue = null;
			try {
				Field field = dtoHelper.field(fieldName);
				boolean accessible = field.isAccessible();
				field.setAccessible(true);
				fieldValue = field.get(object);
				field.setAccessible(accessible);
			} catch (IllegalArgumentException e) {
				throw new DynamicDaoDataAccessException("Illegal argument for field " + fieldName, e);
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access to field " + fieldName, e);
			}
			int sqlType = dtoHelper.column(fieldName).type(); 
			if (sqlType != 0) {
				paramSource.addValue(fieldName + "_optlock", fieldValue, sqlType);
			} else {
				paramSource.addValue(fieldName + "_optlock", fieldValue);				
			}
			whereCounter++;
		}
		
		//execute generators after the optimistic locking fields have been identified
		for (String fieldName : dtoHelper.fieldNames()) {
			Field field = dtoHelper.field(fieldName);
			Column column = field.getAnnotation(Column.class);
			Class<? extends Generator> generatorClass = column.generator();
			if (column.generateOnUpdate() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
				try {
					Generator generator = generators.get(generatorClass);
					if (generator == null) {
						generator = generatorClass.newInstance();
						generators.put(generatorClass, generator);
					}
					boolean accessible = field.isAccessible();
					field.setAccessible(true);
					generator.generateValue(field, object, Generator.Operation.UPDATE);
					field.setAccessible(accessible);
				} catch (InstantiationException e) {
					throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
				} catch (IllegalAccessException e) {
					throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
				}
			}
		}
		
		int updated = namedParameterjdbcTemplate.update(sqlBuilder.toString(), paramSource);
		if (updated == 0 && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Update failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return updated;
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#delete(java.lang.Object)
	 */
	public int delete(Object object) {
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
		Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
		if (primaryKeyFieldNames.size() == 0) {
			throw new DynamicDaoDataAccessException("Unable to uniquely identify object " + object + ", because it has no primary key fields");
		}
		StringBuilder sqlBuilder = new StringBuilder(baseDeleteSQL(object.getClass()));
		sqlBuilder.append(" where ");
		int whereCounter = 0;
		for(String fieldName : primaryKeyFieldNames) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName);
			whereCounter++;
		}
		for(String fieldName : dtoHelper.optimisticLockingFieldNames()) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName + "_optlock");
			Object fieldValue = null;
			try {
				Field field = dtoHelper.field(fieldName);
				boolean accessible = field.isAccessible();
				field.setAccessible(true);
				fieldValue = field.get(object);
				field.setAccessible(accessible);
			} catch (IllegalArgumentException e) {
				throw new DynamicDaoDataAccessException("Illegal argument for field " + fieldName, e);
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access to field " + fieldName, e);
			}
			int sqlType = dtoHelper.column(fieldName).type(); 
			if (sqlType != 0) {
				paramSource.addValue(fieldName + "_optlock", fieldValue, sqlType);
			} else {
				paramSource.addValue(fieldName + "_optlock", fieldValue);				
			}
			whereCounter++;
		}
		
		int updated = namedParameterjdbcTemplate.update(sqlBuilder.toString(), paramSource);
		if (updated == 0 && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Delete failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return updated;
	}

	/**
	 * Generates a select statement for the supplied class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseSelectSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		return "select " + columnList(dtoHelper) + " from " + dtoHelper.table();
	}

	/**
	 * Generates a sql insert statement for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String insertSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		
		StringBuilder sb = new StringBuilder("insert into ");
		sb.append(dtoHelper.table());
		sb.append(" (");
		int columnCount = 0;
		
		StringBuilder fieldBuilder = new StringBuilder();
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
					fieldBuilder.append(", ");
				}
				sb.append(column.value());
				fieldBuilder.append(":").append(fieldName);
				columnCount++;
			}
		}
		sb.append(") values (").append(fieldBuilder.toString()).append(")");
		return sb.toString();
	}

	/**
	 * Generates an update statement without the WHERE clause
	 * for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseUpdateSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sb = new StringBuilder("update ");
		sb.append(dtoHelper.table());
		sb.append(" set ");
		int columnCount = 0;
		
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
				}
				sb.append(column.value()).append(" = :").append(fieldName);
				columnCount++;
			}
		}
		return sb.toString();
	}

	/**
	 * Generates a delete statement without the WHERE clause
	 * for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseDeleteSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sb = new StringBuilder("delete from ");
		sb.append(dtoHelper.table());
		return sb.toString();
	}

	/**
	 * Generates a comma-delimited list of column names based
	 * on the columns provided by the dtoHelper.
	 * 
	 * @param dtoHelper
	 * @return
	 */
	protected String columnList(DTOHelper dtoHelper) {
		StringBuilder sb = new StringBuilder();
		for (Column column : dtoHelper.columns()) {
			if (sb.length() > 0) {
				sb.append(", ");
			}
			sb.append(column.value());			
		}
		return sb.toString();
	}

	/**
	 * Translates a field name into a database column name.
	 * 
	 * @param dtoHelper
	 * @param fieldName
	 * @return
	 */
	protected String resolveColumnName(DTOHelper dtoHelper, String fieldName) {
		return dtoHelper.columnName(fieldName);
	}
}
