package com.jeffknecht.dynadao.impl;


import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.HashMap;

import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.support.lob.DefaultLobHandler;
import org.springframework.jdbc.support.lob.LobHandler;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.api.ColumnMapper;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;

/**
 * Dynamically maps records in a resultset to instances of a class
 * that is annotated with a Table and at least one Column annotation.
 * The class definition must have a default constructor.
 * 
 * @author jeff.knecht
 *
 * @param <T>
 */
public class DynaRowMapper<T> implements RowMapper<T> {
	private Class<? extends T> klazz;
	private LobHandler lobHandler = new DefaultLobHandler();
	private DTOHelper dtoHelper;
	private static HashMap<Class<? extends ColumnMapper>, ColumnMapper> columnMappers = new HashMap<Class<? extends ColumnMapper>, ColumnMapper>();
	
	/**
	 * Instantiates a DynaRowMapper.  The supplied class must
	 * be the same as the type defined for this 
	 * @param klazz
	 */
	public DynaRowMapper(Class<? extends T> klazz) {
		this.klazz = klazz;
		dtoHelper = DTOHelper.get(klazz);
	}
	
	/**
	 * Maps the current row in the resultset to an instance of the
	 * underlying class definition.
	 */
	public T mapRow(ResultSet rs, int arg1) throws SQLException {
		try {
			T t = (T) klazz.newInstance();
			//find the column definitions and get the value out of the resultset
			for (String fieldName : dtoHelper.fieldNames()) {
				Column column = dtoHelper.column(fieldName);
				String columnName = column.value();
				int columnType = column.type();
				boolean trim = column.trim();
				Field f = dtoHelper.field(fieldName);
				Class<?> fieldType = f.getType();
				boolean accessible = f.isAccessible();
				f.setAccessible(true);
				try {
					Class<? extends ColumnMapper> columnMapperClass = column.columnMapper();
					if (columnMapperClass != null && !(columnMapperClass.equals(ColumnMapper.class))) {
						ColumnMapper customMapper = columnMappers.get(columnMapperClass);
						if (customMapper == null) {
							customMapper = column.columnMapper().newInstance();
							columnMappers.put(columnMapperClass, customMapper);
						}
						f.set(t, customMapper.mapValue(rs, columnName));
					} else if (fieldType.equals(String.class)) {
						if (columnType == Types.CLOB) {
							if (trim) {
								f.set(t, trim(lobHandler.getClobAsString(rs, columnName)));
							} else {
								f.set(t, lobHandler.getClobAsString(rs, columnName));
							}
						} else {
							if (trim) {
								f.set(t, trim(rs.getString(columnName)));
							} else {
								f.set(t, rs.getString(columnName));
							}
						}
					} else if (fieldType.equals(int.class)) {
						f.set(t, rs.getInt(columnName));
					} else if (fieldType.equals(Integer.class)) {
						f.set(t, rs.getInt(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(long.class)) {
						f.set(t, rs.getLong(columnName));
					} else if (fieldType.equals(Long.class)) {
						f.set(t, rs.getLong(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(short.class)) {
						f.set(t, rs.getShort(columnName));
					} else if (fieldType.equals(Short.class)) {
						f.set(t, rs.getShort(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(double.class)) {
						f.set(t, rs.getDouble(columnName));
					} else if (fieldType.equals(Double.class)) {
						f.set(t, rs.getDouble(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(float.class)) {
						f.set(t, rs.getFloat(columnName));
					} else if (fieldType.equals(Float.class)) {
						f.set(t, rs.getFloat(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(boolean.class)) {
						f.set(t, rs.getBoolean(columnName));
					} else if (fieldType.equals(Boolean.class)) {
						f.set(t, rs.getBoolean(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(char.class)) {
						String val = rs.getString(columnName);
						f.set(t, val == null || val.length() == 0 ? 0 : val.charAt(0));
					} else if (fieldType.equals(Character.class)) {
						String val = rs.getString(columnName);
						f.set(t, val == null || val.length() == 0 ? null : val.charAt(0));
					} else if (fieldType.equals(java.util.Date.class)) {
						f.set(t, rs.getTimestamp(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(java.sql.Date.class)) {
						f.set(t, rs.getDate(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(Time.class)) {
						f.set(t, rs.getTime(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(Timestamp.class)) {
						f.set(t, rs.getTimestamp(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(byte[].class)) {
						f.set(t, lobHandler.getBlobAsBytes(rs, columnName));
					} else if (fieldType.equals(BigDecimal.class)) {
						f.set(t, rs.getBigDecimal(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else if (fieldType.equals(BigInteger.class)) {
						f.set(t, rs.getLong(columnName));
						if (rs.wasNull()) {
							f.set(t, null);
						}
					} else {
						f.set(t, rs.getObject(columnName));
					}
				} catch (IllegalAccessException e) {
					throw new DynamicDaoDataAccessException("Illegal access on field " + f, e);
				} finally {
					f.setAccessible(accessible);
				}
			}
			return t;
		} catch (InstantiationException e) {
			throw new DynamicDaoDataAccessException("Unable to instantiate class " + klazz, e);
		} catch (IllegalAccessException e) {
			throw new DynamicDaoDataAccessException("Illegal access on class " + klazz, e);
		}
	}

	private String trim(String string) {
		if (string == null) {
			return null;
		} else {
			return string.trim();
		}
	}
}
